import os
import tempfile

import pandas as pd
import pytest

from autorag import generator_models
from autorag.nodes.generator import LlamaIndexLLM
from autorag.nodes.generator.run import run_generator_node
from autorag.utils.util import load_summary_file
from tests.mock import MockLLM

qa_df = pd.DataFrame(
	{
		"qid": ["id-1", "id-2", "id-3"],
		"query": ["query-1", "query-2", "query-3"],
		"generation_gt": [
			["The dog had bit the man.", "The man had bitten the dog."],
			["I want to be a artist, but I end up to be a programmer."],
			[
				"To be a artist these days, you can overcome by AI.",
				"To be a programmer these days, you can overcome by AI.",
				"To be a lawyer these days, you can overcome by AI.",
			],
		],
	}
)

previous_df = pd.DataFrame(
	{
		"qid": ["id-1", "id-2", "id-3"],
		"query": ["query-1", "query-2", "query-3"],
		"prompts": [
			"What was the dog doing with the man?",
			"What is your dream job? And what is your current job?",
			"Is AI can overcome and replace all jobs in the future?",
		],
	}
)


@pytest.fixture
def node_line_dir():
	with tempfile.TemporaryDirectory(ignore_cleanup_errors=True) as temp_dir:
		os.makedirs(os.path.join(temp_dir, "data"))
		qa_df.to_parquet(os.path.join(temp_dir, "data", "qa.parquet"), index=False)
		trial_dir = os.path.join(temp_dir, "test_trial")
		os.makedirs(trial_dir)
		node_line_dir = os.path.join(trial_dir, "test_node_line")
		yield node_line_dir


def test_run_generator_node(node_line_dir):
	generator_models["mock"] = MockLLM
	modules = [LlamaIndexLLM, LlamaIndexLLM]
	module_params = [
		{
			"llm": "mock",
			"temperature": 0.5,
			"top_p": 0.9,
			"max_tokens": 128,
			"batch": 8,
		},
		{"llm": "mock", "temperature": 1.5, "top_p": 0.9, "max_tokens": 32, "batch": 8},
	]
	strategies = {
		"strategy": "rank",
		"metrics": [
			{"metric_name": "bleu"},
			{"metric_name": "meteor"},
			{"metric_name": "rouge"},
		],
		"speed_threshold": 5,
		"token_threshold": 64,
	}
	best_result = run_generator_node(
		modules, module_params, previous_df, node_line_dir, strategies
	)
	assert os.path.exists(os.path.join(node_line_dir, "generator"))
	expect_columns = {
		"qid",
		"query",
		"prompts",
		"generated_texts",
		"generated_tokens",
		"generated_log_probs",
		"bleu",
		"meteor",
		"rouge",
	}
	assert set(best_result.columns) == expect_columns

	summary_path = os.path.join(node_line_dir, "generator", "summary.csv")
	assert os.path.exists(summary_path)
	summary_df = load_summary_file(summary_path)
	expect_columns = {
		"filename",
		"bleu",
		"meteor",
		"rouge",
		"module_name",
		"module_params",
		"execution_time",
		"average_output_token",
		"is_best",
	}
	assert set(summary_df.columns) == expect_columns
	assert len(summary_df) == 2
	assert summary_df["module_params"][0] == {
		"llm": "mock",
		"temperature": 0.5,
		"top_p": 0.9,
		"max_tokens": 128,
		"batch": 8,
	}

	first_path = os.path.join(node_line_dir, "generator", "0.parquet")
	assert os.path.exists(first_path)

	best_path = summary_df[summary_df["is_best"]]["filename"].values[0]
	assert os.path.exists(os.path.join(node_line_dir, "generator", f"best_{best_path}"))



import time
from unittest.mock import patch

from openai import AsyncOpenAI
import openai.resources.beta.chat
from autorag.data.qa.evolve.openai_query_evolve import (
	conditional_evolve_ragas,
	Response,
	reasoning_evolve_ragas,
	compress_ragas,
)
from autorag.data.qa.schema import QA
from tests.autorag.data.qa.evolve.base_test_query_evolve import qa_df
from openai.types.chat import (
	ParsedChatCompletion,
	ParsedChatCompletionMessage,
	ParsedChoice,
)


async def mock_gen_gt_response(*args, **kwargs) -> ParsedChatCompletion[Response]:
	return ParsedChatCompletion(
		id="test_id",
		choices=[
			ParsedChoice(
				finish_reason="stop",
				index=0,
				message=ParsedChatCompletionMessage(
					parsed=Response(evolved_query="mock answer"),
					role="assistant",
				),
			)
		],
		created=int(time.time()),
		model="gpt-4o-mini-2024-07-18",
		object="chat.completion",
	)


client = AsyncOpenAI()


@patch.object(
	openai.resources.beta.chat.completions.AsyncCompletions,
	"parse",
	mock_gen_gt_response,
)
def test_conditional_evolve_ragas():
	qa = QA(qa_df)
	new_qa = qa.batch_apply(conditional_evolve_ragas, client=client)
	assert "query" in new_qa.data.columns
	assert all(isinstance(query, str) for query in new_qa.data["query"].tolist())
	assert len(new_qa.data) == len(qa_df)
	assert all(
		x != y for x, y in zip(new_qa.data["query"].tolist(), qa_df["query"].tolist())
	)


@patch.object(
	openai.resources.beta.chat.completions.AsyncCompletions,
	"parse",
	mock_gen_gt_response,
)
def test_reasoning_evolve_ragas():
	qa = QA(qa_df)
	new_qa = qa.batch_apply(reasoning_evolve_ragas, client=client)
	assert "query" in new_qa.data.columns
	assert all(isinstance(query, str) for query in new_qa.data["query"].tolist())
	assert len(new_qa.data) == len(qa_df)
	assert all(
		x != y for x, y in zip(new_qa.data["query"].tolist(), qa_df["query"].tolist())
	)


@patch.object(
	openai.resources.beta.chat.completions.AsyncCompletions,
	"parse",
	mock_gen_gt_response,
)
def test_compress_ragas():
	qa = QA(qa_df)
	new_qa = qa.batch_apply(compress_ragas, client=client)
	assert "query" in new_qa.data.columns
	assert all(isinstance(query, str) for query in new_qa.data["query"].tolist())
	assert len(new_qa.data) == len(qa_df)
	assert all(
		x != y for x, y in zip(new_qa.data["query"].tolist(), qa_df["query"].tolist())
	)



import os.path
import tempfile

import pytest

from autorag.nodes.promptmaker import WindowReplacement
from tests.autorag.nodes.promptmaker.test_prompt_maker_base import (
	prompt,
	queries,
	retrieved_contents,
	retrieved_metadata,
	previous_result,
	corpus_df,
)


@pytest.fixture
def pseudo_project_dir():
	with tempfile.TemporaryDirectory(ignore_cleanup_errors=True) as project_dir:
		data_dir = os.path.join(project_dir, "data")
		os.makedirs(data_dir)
		corpus_df.to_parquet(os.path.join(data_dir, "corpus.parquet"))
		yield project_dir


@pytest.fixture
def window_replacement_instance(pseudo_project_dir):
	return WindowReplacement(project_dir=pseudo_project_dir)


def test_window_replacement(window_replacement_instance):
	result_prompts = window_replacement_instance._pure(
		prompt, queries, retrieved_contents, retrieved_metadata
	)
	assert len(result_prompts) == 2
	assert isinstance(result_prompts, list)
	assert (
		result_prompts[0]
		== "Answer this question: What is the capital of Japan? \n\n havertz arsenal doosan minji naeun gaeun lets go\n\nhavertz arsenal doosan minji naeun gaeun lets go"
	)
	assert (
		result_prompts[1]
		== "Answer this question: What is the capital of China? \n\n havertz arsenal doosan minji naeun gaeun lets go\n\nhavertz arsenal doosan minji naeun gaeun lets go"
	)


def test_window_replacement_node(pseudo_project_dir):
	result = WindowReplacement.run_evaluator(
		project_dir=pseudo_project_dir, previous_result=previous_result, prompt=prompt
	)
	assert len(result) == 2
	assert result.columns == ["prompts"]
	assert (
		result["prompts"][0]
		== "Answer this question: What is the capital of Japan? \n\n havertz arsenal doosan minji naeun gaeun lets go\n\nhavertz arsenal doosan minji naeun gaeun lets go"
	)
	assert (
		result["prompts"][1]
		== "Answer this question: What is the capital of China? \n\n havertz arsenal doosan minji naeun gaeun lets go\n\nhavertz arsenal doosan minji naeun gaeun lets go"
	)



import time
from unittest.mock import patch

import pandas as pd
from llama_index.llms.openai import OpenAI
from openai import AsyncOpenAI
from openai.types.chat import (
	ParsedChatCompletion,
	ParsedChoice,
	ParsedChatCompletionMessage,
)
import openai.resources.beta.chat

from llama_index.core.base.llms.types import ChatResponse, ChatMessage, MessageRole
from autorag.data.qa.filter.passage_dependency import (
	passage_dependency_filter_openai,
	passage_dependency_filter_llama_index,
	Response,
)
from autorag.data.qa.schema import QA

en_qa_df = pd.DataFrame(
	{
		"query": [
			"What is the most significant discovery mentioned in the research paper?",
			"What was the ruling in the case described in this legal brief?",
			"What is the passage reranker role in the Advanced RAG system?",
			"Who is the latest 30 homerun 30 steal record owner in the KBO league?",
		]
	}
)

ko_qa_df = pd.DataFrame(
	{
		"query": [
			"연구 논문에서 언급된 가장 중요한 발견은 무엇입니까?",
			"이 판결문에 기술된 사건의 판결은 무엇이었습니까?",
			"Advanced RAG 시스템에서 리랭커 역할은 무엇입니까?",
			"KBO 리그에서 가장 최근 30홈런 30도루를 기록한 선수는 누구입니까?",
		]
	}
)

expected_df_en = pd.DataFrame(
	{
		"query": [
			"What is the passage reranker role in the Advanced RAG system?",
			"Who is the latest 30 homerun 30 steal record owner in the KBO league?",
		]
	}
)

expected_df_ko = pd.DataFrame(
	{
		"query": [
			"Advanced RAG 시스템에서 리랭커 역할은 무엇입니까?",
			"KBO 리그에서 가장 최근 30홈런 30도루를 기록한 선수는 누구입니까?",
		]
	}
)

passage_dependent_response = [
	"What is the most significant discovery mentioned in the research paper?",
	"What was the ruling in the case described in this legal brief?",
	"연구 논문에서 언급된 가장 중요한 발견은 무엇입니까?",
	"이 판결문에 기술된 사건의 판결은 무엇이었습니까?",
]


async def mock_openai_response(*args, **kwargs) -> ParsedChatCompletion[Response]:
	user_prompt = kwargs["messages"][1]["content"]
	return ParsedChatCompletion(
		id="test_id",
		choices=[
			ParsedChoice(
				finish_reason="stop",
				index=0,
				message=ParsedChatCompletionMessage(
					parsed=Response(
						is_passage_dependent=user_prompt.split("\n")[0]
						.split(":")[1]
						.strip()
						in passage_dependent_response
					),
					role="assistant",
				),
			)
		],
		created=int(time.time()),
		model="gpt-4o-mini-2024-07-18",
		object="chat.completion",
	)


async def mock_llama_index_response(*args, **kwargs) -> ChatResponse:
	user_prompt = kwargs["messages"][1].content
	return ChatResponse(
		message=ChatMessage(
			role=MessageRole.ASSISTANT,
			content=str(
				user_prompt.split("\n")[0].split(":")[1].strip()
				in passage_dependent_response
			),
		)
	)


@patch.object(
	openai.resources.beta.chat.completions.AsyncCompletions,
	"parse",
	mock_openai_response,
)
def test_passage_dependency_filter_openai():
	client = AsyncOpenAI()
	en_qa = QA(en_qa_df)
	result_en_qa = en_qa.batch_filter(
		passage_dependency_filter_openai, client=client, lang="en"
	).map(lambda df: df.reset_index(drop=True))
	pd.testing.assert_frame_equal(result_en_qa.data, expected_df_en)

	ko_qa = QA(ko_qa_df)
	result_ko_qa = ko_qa.batch_filter(
		passage_dependency_filter_openai, client=client, lang="ko"
	).map(lambda df: df.reset_index(drop=True))
	pd.testing.assert_frame_equal(result_ko_qa.data, expected_df_ko)


@patch.object(
	OpenAI,
	"achat",
	mock_llama_index_response,
)
def test_passage_dependency_filter_llama_index():
	llm = OpenAI(temperature=0, model="gpt-4o-mini")
	en_qa = QA(en_qa_df)
	result_en_qa = en_qa.batch_filter(
		passage_dependency_filter_llama_index, llm=llm, lang="en"
	).map(lambda df: df.reset_index(drop=True))
	pd.testing.assert_frame_equal(result_en_qa.data, expected_df_en)

	ko_qa = QA(ko_qa_df)
	result_ko_qa = ko_qa.batch_filter(
		passage_dependency_filter_llama_index, llm=llm, lang="ko"
	).map(lambda df: df.reset_index(drop=True))
	pd.testing.assert_frame_equal(result_ko_qa.data, expected_df_ko)



import numpy as np
import pytest

from autorag.evaluation.metric import (
	retrieval_f1,
	retrieval_precision,
	retrieval_recall,
	retrieval_ndcg,
	retrieval_mrr,
	retrieval_map,
)
from autorag.schema.metricinput import MetricInput

retrieval_gt = [
	[["test-1", "test-2"], ["test-3"]],
	[["test-4", "test-5"], ["test-6", "test-7"], ["test-8"]],
	[["test-9", "test-10"]],
	[["test-11"], ["test-12"], ["test-13"]],
	[["test-14"]],
	[[]],
	[[""]],
	[["test-15"]],
]

pred = [
	["test-1", "pred-1", "test-2", "pred-3"],  # recall: 0.5, precision: 0.5, f1: 0.5
	["test-6", "pred-5", "pred-6", "pred-7"],  # recall: 1/3, precision: 0.25, f1: 2/7
	["test-9", "pred-0", "pred-8", "pred-9"],  # recall: 1.0, precision: 0.25, f1: 2/5
	[
		"test-13",
		"test-12",
		"pred-10",
		"pred-11",
	],  # recall: 2/3, precision: 0.5, f1: 4/7
	["test-14", "pred-12"],  # recall: 1.0, precision: 0.5, f1: 2/3
	["pred-13"],  # retrieval_gt is empty so not counted
	["pred-14"],  # retrieval_gt is empty so not counted
	["pred-15", "pred-16", "test-15"],  # recall:1, precision: 1/3, f1: 0.5
]
metric_inputs = [MetricInput(retrieval_gt=ret_gt, retrieved_ids=pr) for ret_gt, pr in zip(retrieval_gt, pred)]

def test_retrieval_f1():
	solution = [0.5, 2 / 7, 2 / 5, 4 / 7, 2 / 3, None, None, 0.5]
	result = retrieval_f1(metric_inputs=metric_inputs)
	for gt, res in zip(solution, result):
		assert gt == pytest.approx(res, rel=1e-4)


def test_numpy_retrieval_metric():
	retrieval_gt_np = [[np.array(["test-1", "test-4"])], np.array([["test-2"]])]
	pred_np = np.array([["test-2", "test-3", "test-1"], ["test-5", "test-6", "test-8"]])
	solution = [1.0, 0.0]
	metric_inputs_np = [MetricInput(retrieval_gt=ret_gt_np, retrieved_ids=pr_np) for ret_gt_np, pr_np in
						zip(retrieval_gt_np, pred_np)]
	result = retrieval_recall(metric_inputs=metric_inputs_np)
	for gt, res in zip(solution, result):
		assert gt == pytest.approx(res, rel=1e-4)


def test_retrieval_recall():
	solution = [0.5, 1 / 3, 1, 2 / 3, 1, None, None, 1]
	result = retrieval_recall(metric_inputs=metric_inputs)
	for gt, res in zip(solution, result):
		assert gt == pytest.approx(res, rel=1e-4)


def test_retrieval_precision():
	solution = [0.5, 0.25, 0.25, 0.5, 0.5, None, None, 1 / 3]
	result = retrieval_precision(metric_inputs=metric_inputs)
	for gt, res in zip(solution, result):
		assert gt == pytest.approx(res, rel=1e-4)


def test_retrieval_ndcg():
	solution = [
		0.7039180890341347,
		0.3903800499921017,
		0.6131471927654584,
		0.7653606369886217,
		1,
		None,
		None,
		0.5,
	]
	result = retrieval_ndcg(metric_inputs=metric_inputs)
	for gt, res in zip(solution, result):
		assert gt == pytest.approx(res, rel=1e-4)


def test_retrieval_mrr():
	solution = [1 / 2, 1 / 3, 1, 1 / 2, 1, None, None, 1 / 3]
	result = retrieval_mrr(metric_inputs=metric_inputs)
	for gt, res in zip(solution, result):
		assert gt == pytest.approx(res, rel=1e-4)


def test_retrieval_map():
	solution = [5 / 12, 1 / 3, 1, 1 / 2, 1, None, None, 1 / 3]
	result = retrieval_map(metric_inputs=metric_inputs)
	for gt, res in zip(solution, result):
		assert gt == pytest.approx(res, rel=1e-4)



from unittest.mock import patch

import pandas as pd
import pytest
from llama_index.embeddings.openai import OpenAIEmbedding

from autorag.nodes.retrieval import HybridRRF
from autorag.nodes.retrieval.hybrid_rrf import rrf_pure, hybrid_rrf
from tests.autorag.nodes.retrieval.test_hybrid_base import (
	sample_ids,
	sample_scores,
	previous_result,
	pseudo_project_dir,
)
from tests.mock import mock_get_text_embedding_batch


def test_hybrid_rrf():
	result_id, result_scores = hybrid_rrf(sample_ids, sample_scores, top_k=3, weight=1)
	assert result_id == [["id-3", "id-1", "id-2"], ["id-4", "id-2", "id-3"]]
	assert result_scores[0] == pytest.approx([1.0, (1 / 4) + (1 / 3), (1 / 3)])
	assert result_scores[1] == pytest.approx([1.0, (1 / 4) + (1 / 3), (1 / 3)])


def test_rrf_pure():
	sample_ids = (["id-1", "id-2", "id-3"], ["id-1", "id-4", "id-3"])
	sample_scores = ([1, 3, 5], [4, 2, 6])
	result_id, result_scores = rrf_pure(sample_ids, sample_scores, rrf_k=1, top_k=3)
	assert result_scores == pytest.approx([1.0, (1 / 4) + (1 / 3), (1 / 3)])
	assert result_id == ["id-3", "id-1", "id-2"]


def test_hybrid_rrf_node(pseudo_project_dir):
	modules = {
		"ids": (
			[
				["id-1", "id-2", "id-3"],
				["id-1", "id-2", "id-3"],
				["id-1", "id-2", "id-3"],
			],
			[
				["id-7", "id-8", "id-9"],
				["id-7", "id-8", "id-9"],
				["id-7", "id-8", "id-9"],
			],
		),
		"scores": (
			[[0.1, 0.2, 0.3], [0.1, 0.2, 0.3], [0.1, 0.2, 0.3]],
			[[0.5, 0.6, 0.7], [0.5, 0.6, 0.7], [0.5, 0.6, 0.7]],
		),
		"top_k": 3,
		"weight": 1,
	}
	result_df = HybridRRF.run_evaluator(
		project_dir=pseudo_project_dir, previous_result=previous_result, **modules
	)
	assert len(result_df) == 3
	assert isinstance(result_df, pd.DataFrame)
	assert set(result_df.columns) == {
		"retrieved_contents",
		"retrieved_ids",
		"retrieve_scores",
	}
	assert set(result_df["retrieved_ids"].tolist()[0]) == {"id-9", "id-3", "id-2"}
	assert result_df["retrieve_scores"].tolist()[0] == pytest.approx([0.5, 0.5, 1 / 3])
	assert set(result_df["retrieved_contents"].tolist()[0]) == {
		"doc-9",
		"doc-3",
		"doc-2",
	}


@patch.object(
	OpenAIEmbedding,
	"get_text_embedding_batch",
	mock_get_text_embedding_batch,
)
def test_hybrid_rrf_node_deploy(pseudo_project_dir):
	modules = {
		"target_modules": ("bm25", "vectordb"),
		"target_module_params": [
			{"top_k": 3},
			{"embedding_model": "openai", "top_k": 3},
		],
		"top_k": 3,
		"weight": 1,
	}
	hybrid_rrf = HybridRRF(project_dir=pseudo_project_dir, **modules)
	result_df = hybrid_rrf.pure(previous_result=previous_result, **modules)
	assert len(result_df) == 3
	assert isinstance(result_df, pd.DataFrame)
	assert set(result_df.columns) == {
		"retrieved_contents",
		"retrieved_ids",
		"retrieve_scores",
	}
	assert len(result_df["retrieved_ids"].tolist()[0]) == 3
	assert len(result_df["retrieve_scores"].tolist()[0]) == 3
	assert len(result_df["retrieved_contents"].tolist()[0]) == 3



import asyncio
import itertools
import os
import pathlib
import tempfile
from datetime import datetime, date

import numpy as np
import pandas as pd
import pytest
import tiktoken
from llama_index.core.base.embeddings.base import BaseEmbedding
from llama_index.core.llms import CompletionResponse
from llama_index.embeddings.openai import OpenAIEmbedding

from autorag.utils import fetch_contents
from autorag.utils.util import (
	load_summary_file,
	result_to_dataframe,
	make_combinations,
	explode,
	replace_value_in_dict,
	normalize_string,
	convert_string_to_tuple_in_dict,
	process_batch,
	convert_env_in_dict,
	openai_truncate_by_token,
	convert_datetime_string,
	split_dataframe,
	find_trial_dir,
	find_node_summary_files,
	normalize_unicode,
	dict_to_markdown,
	dict_to_markdown_table,
	convert_inputs_to_list,
	to_list,
	get_event_loop,
	find_key_values,
	pop_params,
	apply_recursive,
)
from tests.mock import MockLLM

root_dir = pathlib.PurePath(os.path.dirname(os.path.realpath(__file__))).parent.parent

summary_df = pd.DataFrame(
	{
		"best_module_name": ["bm25", "upr", "gpt-4"],
		"best_module_params": [
			{"top_k": 50},
			{"model": "llama-2", "havertz": "chelsea"},
			{"top_p": 0.9},
		],
	}
)


@pytest.fixture
def module_name():
	return "test_module"


@pytest.fixture
def module_params():
	return {
		"param1": "value1",
		"param2": "value2",
		"param3": "value3",
	}


@pytest.fixture
def summary_path():
	with tempfile.TemporaryDirectory(ignore_cleanup_errors=True) as tmp_dir:
		summary_path = os.path.join(tmp_dir, "summary.csv")
		summary_df.to_csv(summary_path, index=False)
		yield summary_path


def test_fetch_contents():
	corpus_data_path = os.path.join(root_dir, "resources", "corpus_data_sample.parquet")
	corpus_data = pd.read_parquet(corpus_data_path)
	search_rows = corpus_data.sample(n=10)
	find_contents = fetch_contents(
		corpus_data, list(map(lambda x: [x], search_rows["doc_id"].tolist()))
	)
	assert len(find_contents) == len(search_rows)
	assert (
		list(itertools.chain.from_iterable(find_contents))
		== search_rows["contents"].tolist()
	)

	corpus_data = pd.DataFrame(
		{
			"doc_id": ["doc1", "doc2", "doc3"],
			"contents": ["apple", "banana", "cherry"],
			"metadata": [
				{"last_modified_datetime": datetime(2022, 1, 1, 0, 0, 0)},
				{"last_modified_datetime": datetime(2022, 1, 2, 0, 0, 0)},
				{"last_modified_datetime": datetime(2022, 1, 3, 0, 0, 0)},
			],
		}
	)
	find_contents = fetch_contents(corpus_data, [["doc3", "doc1"], ["doc2"]])
	assert find_contents[0] == ["cherry", "apple"]
	assert find_contents[1] == ["banana"]

	find_metadatas = fetch_contents(
		corpus_data, [["doc3", "doc1"], ["doc2"]], "metadata"
	)
	assert find_metadatas[0] == [
		{"last_modified_datetime": datetime(2022, 1, 3, 0, 0, 0)},
		{"last_modified_datetime": datetime(2022, 1, 1, 0, 0, 0)},
	]
	assert find_metadatas[1] == [
		{"last_modified_datetime": datetime(2022, 1, 2, 0, 0, 0)}
	]

	find_empty = fetch_contents(corpus_data, [[], ["doc2"]])
	assert find_empty[0] == [None]
	assert find_empty[1] == ["banana"]

	find_blank = fetch_contents(corpus_data, [[""], ["doc2"]])
	assert find_blank[0] == [None]
	assert find_blank[1] == ["banana"]


def test_load_summary_file(summary_path):
	with pytest.raises(ValueError):
		load_summary_file(summary_path)
	df = load_summary_file(summary_path, ["best_module_params"])
	assert df.equals(summary_df)


def test_load_summary_file_recency_filter():
	df = pd.DataFrame(
		{
			"module_name": ["havertz", "recency_filter"],
			"module_params": [
				{"jazz": "eastsidegunn"},
				{"threshold": datetime(2022, 1, 3, 0, 1, 3)},
			],
		}
	)
	with tempfile.NamedTemporaryFile(suffix=".csv", delete=False) as csv_file:
		df.to_csv(csv_file.name, index=False)
		load_df = load_summary_file(csv_file.name)
		assert load_df.equals(df)
		csv_file.close()
		os.unlink(csv_file.name)


def test_convert_datetime_string():
	datetime_dict = {"threshold": datetime(2022, 1, 3, 0, 0, 3)}
	date_dict = {"threshold": date(2001, 7, 11)}
	result1 = convert_datetime_string(str(datetime_dict))
	result2 = convert_datetime_string(str(date_dict))

	assert result1 == datetime_dict["threshold"]
	assert result2 == date_dict["threshold"]


def test_result_to_dataframe():
	@result_to_dataframe(["col_1", "col_2"])
	def func1():
		return [1, 2], [3, 4]

	result1 = func1()
	assert isinstance(result1, pd.DataFrame)
	assert result1.columns.tolist() == ["col_1", "col_2"]
	assert result1["col_1"].tolist() == [1, 2]
	assert result1["col_2"].tolist() == [3, 4]

	@result_to_dataframe(["col_1"])
	def func2():
		return [1, 2, 3]

	result2 = func2()
	assert isinstance(result2, pd.DataFrame)
	assert result2.columns.tolist() == ["col_1"]
	assert result2["col_1"].tolist() == [1, 2, 3]


def test_make_combinations():
	target_dict = {
		"key1": "value1",
		"key2": ["value1", "value2"],
		"key3": "value3",
		"key4": ["value4", "value5"],
	}
	solution = [
		{"key1": "value1", "key2": "value1", "key3": "value3", "key4": "value4"},
		{"key1": "value1", "key2": "value1", "key3": "value3", "key4": "value5"},
		{"key1": "value1", "key2": "value2", "key3": "value3", "key4": "value4"},
		{"key1": "value1", "key2": "value2", "key3": "value3", "key4": "value5"},
	]
	combinations = make_combinations(target_dict)
	assert len(combinations) == len(solution)
	assert all([combination in solution for combination in combinations])

	elem1 = {"key5": "value5", "key6": ["value6", "value7"]}
	elem2 = {"key7": "value8"}
	value_of_key_4 = [elem1, elem2]
	target_dict = {
		"key1": "value1",
		"key2": ["value1", "value2"],
		"key3": "value3",
		"key4": value_of_key_4,
	}
	combinations = make_combinations(target_dict)
	solution = [
		{"key1": "value1", "key2": "value1", "key3": "value3", "key4": elem1},
		{"key1": "value1", "key2": "value2", "key3": "value3", "key4": elem1},
		{"key1": "value1", "key2": "value1", "key3": "value3", "key4": elem2},
		{"key1": "value1", "key2": "value2", "key3": "value3", "key4": elem2},
	]
	assert len(combinations) == len(solution)
	assert all([combination in solution for combination in combinations])

	target_dict = {
		"key1": "value1",
		"key2": ["value1", "value2"],
		"key3": "value3",
		"key4": ("value4", "value5"),
	}
	solution = [
		{
			"key1": "value1",
			"key2": "value1",
			"key3": "value3",
			"key4": ("value4", "value5"),
		},
		{
			"key1": "value1",
			"key2": "value2",
			"key3": "value3",
			"key4": ("value4", "value5"),
		},
	]
	combinations = make_combinations(target_dict)
	assert len(combinations) == len(solution)
	assert all([combination in solution for combination in combinations])


def test_explode():
	index_values = ["a", "b", "c"]
	explode_values = [
		["apple", "banana", "cherry"],
		["april", "may"],
		["alpha"],
	]
	result_index, result_values = explode(index_values, explode_values)
	assert result_index == ["a", "a", "a", "b", "b", "c"]
	assert result_values == ["apple", "banana", "cherry", "april", "may", "alpha"]


def test_replace_value_in_dict():
	target_dict = {
		"key1": "value1",
		"key2": "value2",
		"key3": "value3",
	}
	result_dict = replace_value_in_dict(target_dict, "key1", "value4")
	assert result_dict == {
		"key1": "value4",
		"key2": "value2",
		"key3": "value3",
	}
	result_dict = replace_value_in_dict(target_dict, "key4", "value4")
	assert result_dict == target_dict


def test_normalize_string():
	text = "This IS a TEST Text."
	expected = "this is test text"
	assert normalize_string(text) == expected

	text = "Hello, world! This is a test."
	expected = "hello world this is test"
	assert normalize_string(text) == expected

	text = "The quick brown fox jumps over the lazy dog."
	expected = "quick brown fox jumps over lazy dog"
	assert normalize_string(text) == expected

	text = "This    is      a test    text."
	expected = "this is test text"
	assert normalize_string(text) == expected

	text = "The, QUICK Brown-Fox; jumps over... the LAZY dog!"
	expected = "quick brownfox jumps over lazy dog"
	assert normalize_string(text) == expected


def test_convert_string_to_tuple_in_dict():
	# Example usage
	data = {
		"key1": "(1, 'two', 3)",
		"key2": ["(4, 5, 'six')", {"nested_key": "(7, 8, 'nine')"}, {"key4": "value2"}],
		"key3": {
			"nested_key2": "(10, 'eleven', 12)",
			"nested_key3": "value1",
			"nested_key4": {"nested_key5": "('thirteen', 14, 15)"},
		},
	}
	result = convert_string_to_tuple_in_dict(data)
	assert result == {
		"key1": (1, "two", 3),
		"key2": [(4, 5, "six"), {"nested_key": (7, 8, "nine")}, {"key4": "value2"}],
		"key3": {
			"nested_key2": (10, "eleven", 12),
			"nested_key3": "value1",
			"nested_key4": {"nested_key5": ("thirteen", 14, 15)},
		},
	}


def test_convert_env_in_dict():
	os.environ["ENV_VAR1"] = "value1"
	os.environ["ENV_VAR2"] = "value2"
	os.environ["ENV_VAR3"] = "value3"
	data = {
		"key1": "value1",
		"key2": [
			"value1",
			"${ENV_VAR1}",
		],
		"key3": "${ENV_VAR2}",
		"key4": {
			"key5": "value1",
			"key6": "${ENV_VAR3}",
			"key7": [
				"value1",
				"${ENV_VAR4}",
			],
		},
		"prompt": "This is a prompt with ${ENV_VAR1} and ${ENV_VAR2}.",
	}
	result = convert_env_in_dict(data)
	assert result == {
		"key1": "value1",
		"key2": [
			"value1",
			"value1",
		],
		"key3": "value2",
		"key4": {
			"key5": "value1",
			"key6": "value3",
			"key7": [
				"value1",
				"",
			],
		},
		"prompt": "This is a prompt with value1 and value2.",
	}


def test_process_batch():
	prompts = [str(i) for i in range(1000)]
	results = [CompletionResponse(text=prompt) for prompt in prompts]
	mock_llm = MockLLM()

	tasks = [mock_llm.acomplete(prompt) for prompt in prompts]
	loop = get_event_loop()
	result = loop.run_until_complete(process_batch(tasks, batch_size=64))

	assert result == results


def test_openai_truncate_by_token():
	base_text = "This is a test text."
	t1 = base_text * 5
	t2 = base_text * 2000
	t3 = base_text * 20

	truncated = openai_truncate_by_token([t1, t2, t3], 8192, "text-embedding-ada-002")
	assert len(truncated) == 3
	assert truncated[0] == base_text * 5
	assert len(truncated[1]) < len(t2)
	assert (
		len(tiktoken.encoding_for_model("text-embedding-ada-002").encode(truncated[1]))
		== 8192
	)
	assert len(truncated[2]) == len(t3)


def test_split_dataframe():
	df = pd.DataFrame({"a": list(range(10)), "b": list(range(10, 20))})

	df_list_1 = split_dataframe(df, chunk_size=5)
	assert len(df_list_1) == 2
	assert len(df_list_1[0]) == 5
	assert pd.DataFrame({"a": list(range(5)), "b": list(range(10, 15))}).equals(
		df_list_1[0]
	)

	df_list_2 = split_dataframe(df, chunk_size=3)
	assert len(df_list_2) == 4
	assert len(df_list_2[0]) == 3
	assert len(df_list_2[-1]) == 1
	assert pd.DataFrame({"a": list(range(3)), "b": list(range(10, 13))}).equals(
		df_list_2[0]
	)


def test_find_trial_dir():
	project_dir = os.path.join(root_dir, "resources", "result_project")
	trial_dirs = find_trial_dir(project_dir)

	assert len(trial_dirs) == 4
	assert all(isinstance(int(os.path.basename(path)), int) for path in trial_dirs)


def test_find_node_summary_files():
	trial_dir = os.path.join(root_dir, "resources", "result_project", "2")
	node_summary_paths = find_node_summary_files(trial_dir)

	assert len(node_summary_paths) == 4
	assert all(os.path.basename(path) == "summary.csv" for path in node_summary_paths)


def test_normalize_unicode():
	str1 = "전국보행자전용도로표준데이터"
	str2 = "전국보행자전용도로표준데이터"
	assert len(str1) == 14
	assert len(str2) == 34
	assert str1 != str2

	new_str1 = normalize_unicode(str1)
	new_str2 = normalize_unicode(str2)

	assert len(new_str1) == 14
	assert len(new_str2) == 14
	assert new_str1 == new_str2


def test_dict_to_markdown():
	data = {
		"Title": "Sample Document",
		"Author": "John Doe",
		"Content": {
			"Introduction": "This is the introduction.",
			"Body": [
				"First point",
				"Second point",
				{"Subsection": "Details about the second point"},
			],
			"Conclusion": "This is the conclusion.",
		},
	}
	markdown_text = dict_to_markdown(data)
	result_text = """# Title
Sample Document
# Author
John Doe
# Content
## Introduction
This is the introduction.
## Body
- First point
- Second point
### Subsection
Details about the second point
## Conclusion
This is the conclusion.
"""
	assert result_text == markdown_text


def test_dict_to_markdown_table():
	data = {"key1": "value1", "key2": "value2"}
	result = dict_to_markdown_table(data, "havertz", "william")
	result_text = """| havertz | william |
| :---: | :-----: |
| key1 | value1 |
| key2 | value2 |
"""
	assert result == result_text


@convert_inputs_to_list
def convert_inputs_to_list_function(int_type, str_type, iterable_type, iterable_type2):
	assert isinstance(int_type, int)
	assert isinstance(str_type, str)
	assert isinstance(iterable_type, list)
	assert isinstance(iterable_type2, list)


def test_convert_inputs_to_list():
	convert_inputs_to_list_function(1, "jax", (2, 3), (5, 6, [4, 66]))
	convert_inputs_to_list_function(
		1, "jax", np.array([3, 4]), [pd.Series([12, 13]), 14]
	)
	convert_inputs_to_list_function(
		4, "jax", pd.Series([7, 8, 9]), np.array([[3, 4], [4, 5]])
	)


def test_to_list():
	embedding_model = OpenAIEmbedding()
	new_model = to_list(embedding_model)
	assert isinstance(new_model, BaseEmbedding)


def test_find_key_values():
	# Test case 1: Simple dictionary
	data = {"a": 1, "b": 2, "c": {"a": 3, "d": 4}}
	target_key = "a"
	expected = [1, 3]
	assert find_key_values(data, target_key) == expected

	# Test case 2: Nested dictionary with lists
	data = {"a": 1, "b": [{"a": 2}, {"c": {"a": 3}}]}
	target_key = "a"
	expected = [1, 2, 3]
	assert find_key_values(data, target_key) == expected

	# Test case 3: List of dictionaries
	data = [{"a": 1}, {"b": 2, "c": {"a": 3}}]
	target_key = "a"
	expected = [1, 3]
	assert find_key_values(data, target_key) == expected

	# Test case 4: No matching key
	data = {"b": 2, "c": {"d": 4}}
	target_key = "a"
	expected = []
	assert find_key_values(data, target_key) == expected

	# Test case 5: Empty dictionary
	data = {}
	target_key = "a"
	expected = []
	assert find_key_values(data, target_key) == expected

	# Test case 6: Empty list
	data = []
	target_key = "a"
	expected = []
	assert find_key_values(data, target_key) == expected

	# Test case 7: Complex nested structure
	data = {"a": {"b": [{"a": 1}, {"c": {"a": 2}}]}, "d": [{"e": {"a": 3}}]}
	target_key = "a"
	expected = [{"b": [{"a": 1}, {"c": {"a": 2}}]}, 1, 2, 3]
	assert find_key_values(data, target_key) == expected


def test_pop_params():
	# Test function with no self or cls
	def func_no_self_cls(param1, param2):
		pass

	kwargs = {"param1": "value1", "param2": "value2", "extra_param": "extra_value"}
	expected = {"param1": "value1", "param2": "value2"}
	result = pop_params(func_no_self_cls, kwargs)
	assert result == expected
	assert kwargs == {"extra_param": "extra_value"}

	# Test function with self
	class TestClass:
		def func_with_self(self, param1, param2):
			pass

	kwargs = {
		"param1": "value1",
		"param2": "value2",
		"extra_param": "extra_value",
		"self": "self",
	}
	expected = {"param1": "value1", "param2": "value2"}
	result = pop_params(TestClass.func_with_self, kwargs)
	assert result == expected
	assert kwargs == {"extra_param": "extra_value", "self": "self"}

	# Test function with cls
	class TestClassWithCls:
		@classmethod
		def func_with_cls(cls, param1, param2):
			pass

	kwargs = {
		"param1": "value1",
		"param2": "value2",
		"extra_param": "extra_value",
		"cls": "cls",
	}
	expected = {"param1": "value1", "param2": "value2"}
	result = pop_params(TestClassWithCls.func_with_cls, kwargs)
	assert result == expected
	assert kwargs == {"extra_param": "extra_value", "cls": "cls"}

	# Test function with no parameters
	def func_no_params():
		pass

	kwargs = {"extra_param": "extra_value"}
	expected = {}
	result = pop_params(func_no_params, kwargs)
	assert result == expected
	assert kwargs == {"extra_param": "extra_value"}

	# Test function with mixed parameters
	def func_mixed(param1, param2, *args, **kwargs):
		pass

	kwargs = {"param1": "value1", "param2": "value2", "extra_param": "extra_value"}
	expected = {"param1": "value1", "param2": "value2"}
	result = pop_params(func_mixed, kwargs)
	assert result == expected
	assert kwargs == {"extra_param": "extra_value"}


def test_apply_recursive():
	data = [1, 2, 3, 4]
	result = apply_recursive(lambda x: x * 2, data)
	assert result == [2, 4, 6, 8]

	data = [[1, 2], [3, 4]]
	result = apply_recursive(lambda x: x * 2, data)
	assert result == [[2, 4], [6, 8]]

	data = [[1, [2, 3]], [4, [5, 6]]]
	result = apply_recursive(lambda x: x * 2, data)
	assert result == [[2, [4, 6]], [8, [10, 12]]]

	data = []
	result = apply_recursive(lambda x: x * 2, data)
	assert result == []

	data = 5
	result = apply_recursive(lambda x: x * 2, data)
	assert result == 10

	data = [(4, 5), (6, 7), [5, [6, 7]], np.array([4, 5]), pd.Series([4, 5])]
	result = apply_recursive(lambda x: x * 2, data)
	assert result == [[8, 10], [12, 14], [10, [12, 14]], [8, 10], [8, 10]]



import pandas as pd
import pytest

from autorag.schema.metricinput import MetricInput


@pytest.fixture
def sample_metric_input():
    return MetricInput(
        query="test query",
        queries=["q1", "q2"],
        retrieval_gt_contents=[["gc1"], ["gc2"]]
    )


def test_default_constructor():
    metric_input = MetricInput()
    assert metric_input.query is None
    assert metric_input.queries is None
    assert metric_input.retrieval_gt_contents is None


def test_is_fields_notnone(sample_metric_input):
    assert sample_metric_input.is_fields_notnone(["query", "queries", "retrieval_gt_contents"])

    sample_metric_input.query = None
    assert not sample_metric_input.is_fields_notnone(["query", "queries", "retrieval_gt_contents"])


def test_from_dataframe_multiple_rows():
    df = pd.DataFrame({
        'query': ['test query 1', 'test query 2'],
        'queries': [['q1', 'q2'], ['q3', 'q4']],
        'retrieval_gt_contents': [['gc1', 'gc2'], ['gc3', 'gc4']]
    })

    metricinputs = MetricInput.from_dataframe(df)
    assert len(metricinputs) == 2
    assert metricinputs[0].query == 'test query 1'
    assert metricinputs[1].query == 'test query 2'


def test_from_dataframe_empty_values():
    df = pd.DataFrame({
        'query': ['', 'valid query'],
        'queries': [[], ['q1', 'q2']],
        'retrieval_gt_contents': [None, ['gc1', 'gc2']]
    })

    metricinputs = MetricInput.from_dataframe(df)
    assert len(metricinputs) == 2
    assert metricinputs[0].query is None
    assert metricinputs[0].queries is None
    assert metricinputs[0].retrieval_gt_contents is None
    assert metricinputs[1].query == 'valid query'
    assert metricinputs[1].queries == ['q1', 'q2']
    assert metricinputs[1].retrieval_gt_contents == ['gc1', 'gc2']



from unittest.mock import patch

import pytest
from llama_index.embeddings.openai import OpenAIEmbedding

from autorag.evaluation.metric import bleu, meteor, rouge, sem_score, g_eval, bert_score
from autorag.schema.metricinput import MetricInput
from tests.delete_tests import is_github_action
from tests.mock import mock_get_text_embedding_batch

generation_gts = [
	["The dog had bit the man.", "The man had bitten the dog."],
	["I want to be a artist, but I end up to be a programmer."],
	[
		"To be a artist these days, you can overcome by AI.",
		"To be a programmer these days, you can overcome by AI.",
		"To be a lawyer these days, you can overcome by AI.",
	],
]
generations = [
	"The dog bit the man.",
	"It really like to be a programmer, but I think artist is my passion.",
	"To be a artist these days, you can overcome by AI.",
]

ko_generation_gts = [
	["개가 남자를 물었다.", "남자가 개를 물었다."],
	["나는 예술가가 되고 싶었지만, 결국 프로그래머가 되었다."],
	[
		"요즘 예술가가 되려면, AI를 이겨야 한다.",
		"요즘 프로그래머가 되려면, AI를 이겨야 한다.",
		"요즘 변호사가 되려면, AI를 이겨야 한다.",
	],
]

ko_generations = [
	"개가 남자를 물었다.",
	"나는 정말이지 예술가가 되고 싶었지만, 결국 프로그래머가 되었다.",
	"요즘 세상에서는 예술가가 되려면, AI를 이겨야 한다.",
]

metric_inputs = [MetricInput(generated_texts=gen, generation_gt=gen_gt) for gen, gen_gt in
				 zip(generations, generation_gts)]
ko_metric_inputs = [MetricInput(generated_texts=gen, generation_gt=gen_gt) for gen, gen_gt in
					zip(ko_generations, ko_generation_gts)]
def base_test_generation_metrics(func, solution, **kwargs):
	scores = func(metric_inputs, **kwargs)
	assert len(scores) == len(generation_gts)
	assert all(isinstance(score, float) for score in scores)
	assert all(
		list(map(lambda x: x[0] == pytest.approx(x[1], 0.001), zip(scores, solution)))
	)


def ko_base_test_generation_metrics(func, solution, **kwargs):
	scores = func(ko_metric_inputs, **kwargs)
	assert len(scores) == len(ko_generation_gts)
	assert all(isinstance(score, float) for score in scores)
	assert all(
		list(map(lambda x: x[0] == pytest.approx(x[1], 0.001), zip(scores, solution)))
	)


def test_bleu():
	base_test_generation_metrics(bleu, [51.1507, 23.5783, 100.0], lowercase=True)


def test_meteor():
	base_test_generation_metrics(
		meteor, [0.454033, 0.2985435, 0.64077828], alpha=0.85, beta=0.2, gamma=0.6
	)


def test_rouge():
	base_test_generation_metrics(rouge, [0.909, 0.35714, 1.0])


@pytest.mark.skipif(
	is_github_action(),
	reason="Skipping this test on GitHub Actions. It use local model",
)
def test_sem_score():
	base_test_generation_metrics(sem_score, [0.9005998, 0.7952, 1.0])


@patch.object(
	OpenAIEmbedding,
	"get_text_embedding_batch",
	mock_get_text_embedding_batch,
)
def test_sem_score_other_model():
	scores = sem_score(
		metric_inputs=metric_inputs,
		embedding_model=OpenAIEmbedding(),
	)
	assert len(scores) == len(generation_gts)
	assert all(isinstance(score, float) for score in scores)


@pytest.mark.skipif(is_github_action(), reason="Skipping this test on GitHub Actions")
def test_g_eval_fluency():
	base_test_generation_metrics(
		g_eval, [3.0, 2.0, 2.0], metrics=["fluency"], model="gpt-3.5-turbo"
	)


@pytest.mark.skipif(is_github_action(), reason="Skipping this test on GitHub Actions")
def test_g_eval_full():
	base_test_generation_metrics(g_eval, [3.5, 2.75, 2.0], model="gpt-4o-mini")


@pytest.mark.skipif(is_github_action(), reason="Skipping this test on GitHub Actions")
def test_bert_score_en():
	base_test_generation_metrics(bert_score, [0.981902, 0.93164, 1.0], n_threads=8)


@pytest.mark.skipif(is_github_action(), reason="Skipping this test on GitHub Actions")
def test_bert_score_ko():
	ko_base_test_generation_metrics(bert_score, [1.0, 0.965312, 0.96309], lang="ko")



from typing import List, Tuple

import numpy as np
import pandas as pd
import torch
from tqdm import tqdm
from transformers import AutoModelForSequenceClassification, AutoTokenizer

from autorag.nodes.passagereranker.base import BasePassageReranker
from autorag.utils.util import (
	make_batch,
	sort_by_scores,
	flatten_apply,
	select_top_k,
	result_to_dataframe,
)


class KoReranker(BasePassageReranker):
	def __init__(self, project_dir: str, *args, **kwargs):
		super().__init__(project_dir)

		model_path = "Dongjin-kr/ko-reranker"
		self.tokenizer = AutoTokenizer.from_pretrained(model_path)
		self.model = AutoModelForSequenceClassification.from_pretrained(model_path)
		self.model.eval()
		# Determine the device to run the model on (GPU if available, otherwise CPU)
		self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
		self.model.to(self.device)

	def __del__(self):
		del self.model
		if torch.cuda.is_available():
			torch.cuda.empty_cache()
		super().__del__()

	@result_to_dataframe(["retrieved_contents", "retrieved_ids", "retrieve_scores"])
	def pure(self, previous_result: pd.DataFrame, *args, **kwargs):
		queries, contents, _, ids = self.cast_to_run(previous_result)
		top_k = kwargs.pop("top_k")
		batch = kwargs.pop("batch", 64)
		return self._pure(queries, contents, ids, top_k, batch)

	def _pure(
		self,
		queries: List[str],
		contents_list: List[List[str]],
		ids_list: List[List[str]],
		top_k: int,
		batch: int = 64,
	) -> Tuple[List[List[str]], List[List[str]], List[List[float]]]:
		"""
		Rerank a list of contents based on their relevance to a query using ko-reranker.
		ko-reranker is a reranker based on korean (https://huggingface.co/Dongjin-kr/ko-reranker).

		:param queries: The list of queries to use for reranking
		:param contents_list: The list of lists of contents to rerank
		:param ids_list: The list of lists of ids retrieved from the initial ranking
		:param top_k: The number of passages to be retrieved
		:param batch: The number of queries to be processed in a batch
		    Default is 64.
		:return: Tuple of lists containing the reranked contents, ids, and scores
		"""
		nested_list = [
			list(map(lambda x: [query, x], content_list))
			for query, content_list in zip(queries, contents_list)
		]
		scores_nps = flatten_apply(
			koreranker_run_model,
			nested_list,
			model=self.model,
			batch_size=batch,
			tokenizer=self.tokenizer,
			device=self.device,
		)

		rerank_scores = list(
			map(
				lambda scores: exp_normalize(np.array(scores)).astype(float), scores_nps
			)
		)

		df = pd.DataFrame(
			{
				"contents": contents_list,
				"ids": ids_list,
				"scores": rerank_scores,
			}
		)
		df[["contents", "ids", "scores"]] = df.apply(
			sort_by_scores, axis=1, result_type="expand"
		)
		results = select_top_k(df, ["contents", "ids", "scores"], top_k)

		return (
			results["contents"].tolist(),
			results["ids"].tolist(),
			results["scores"].tolist(),
		)


def koreranker_run_model(input_texts, model, tokenizer, device, batch_size: int):
	batch_input_texts = make_batch(input_texts, batch_size)
	results = []
	for batch_texts in tqdm(batch_input_texts):
		inputs = tokenizer(
			batch_texts,
			padding=True,
			truncation=True,
			return_tensors="pt",
			max_length=512,
		)
		inputs = inputs.to(device)
		with torch.no_grad():
			scores = (
				model(**inputs, return_dict=True)
				.logits.view(
					-1,
				)
				.float()
			)
			scores_np = scores.cpu().numpy()
			results.extend(scores_np)
	return results


def exp_normalize(x):
	b = x.max()
	y = np.exp(x - b)
	return y / y.sum()



import os
import tempfile
from unittest.mock import patch

import numpy as np
import pandas as pd
import pytest
from llama_index.core.base.llms.types import CompletionResponse
from llama_index.llms.openai import OpenAI

from autorag import generator_models
from autorag.evaluation.util import cast_metrics
from autorag.nodes.generator import LlamaIndexLLM
from autorag.nodes.promptmaker import Fstring
from autorag.nodes.promptmaker.run import (
	evaluate_generator_result,
	evaluate_one_prompt_maker_node,
	run_prompt_maker_node,
)
from autorag.schema.metricinput import MetricInput
from autorag.utils.util import load_summary_file
from tests.mock import MockLLM

prompts = [
	"Hello, Do you know the world without war?",
	"Hi, I am dreaming about the world without any war.",
]
sample_generated_texts = [
	"hello. This is the world speaking",
	"Hello, Do you know the world without war?",
]
sample_generation_gt = [
	["Hello from space. Hi! This is the world speaking."],
	[
		"Hello, Do you know the world without war?",
		"Hi, I am dreaming about the world without any war.",
	],
]
metrics = [{"metric_name": "bleu"}, {"metric_name": "rouge"}]
previous_result = pd.DataFrame(
	{
		"query": ["What is the war?", "Who is John Lennon?"],
		"retrieved_contents": [
			["War is horrible.", "War is bad."],
			["John Lennon is a musician.", "John Lennon is a singer."],
		],
		"test_column": ["test_value_1", "test_value_2"],
	}
)


def test_evaluate_generator_result():
	sample_df = pd.DataFrame({"generated_texts": sample_generated_texts})
	metric_inputs = [
		MetricInput(generation_gt=gen_gt) for gen_gt in sample_generation_gt
	]
	result_df = evaluate_generator_result(sample_df, metric_inputs, metrics)
	metric_names, _ = cast_metrics(metrics)
	assert all(metric_name in result_df.columns for metric_name in metric_names)
	assert len(result_df) == len(sample_generated_texts)


def test_evaluate_one_prompt_maker_node():
	generator_models["mock"] = MockLLM
	generator_funcs = [LlamaIndexLLM, LlamaIndexLLM]
	generator_params = [
		{"llm": "mock", "model": "gpt-3.5-turbo"},
		{"llm": "mock", "model": "gpt-3.5-turbo-1106"},
	]
	project_dir = "_"
	metric_inputs = [
		MetricInput(generation_gt=gen_gt) for gen_gt in sample_generation_gt
	]
	best_result = evaluate_one_prompt_maker_node(
		prompts,
		generator_funcs,
		generator_params,
		metric_inputs,
		metrics,
		project_dir,
		"rank",
	)
	metric_names, _ = cast_metrics(metrics)
	assert isinstance(best_result, pd.DataFrame)
	assert all(metric_name in best_result.columns for metric_name in metric_names)
	assert len(best_result) == len(prompts)


@pytest.fixture
def node_line_dir():
	with tempfile.TemporaryDirectory(ignore_cleanup_errors=True) as project_dir:
		data_dir = os.path.join(project_dir, "data")
		os.makedirs(data_dir)
		qa_data = pd.DataFrame(
			{
				"qid": ["id-1", "id-2"],
				"query": ["What is the war?", "Who is John Lennon?"],
				"retrieval_gt": [[["doc-1"]], [["doc-2"]]],
				"generation_gt": sample_generation_gt,
			}
		)
		qa_data.to_parquet(os.path.join(data_dir, "qa.parquet"), index=False)
		trial_dir = os.path.join(project_dir, "trial")
		os.makedirs(trial_dir)
		node_line_path = os.path.join(trial_dir, "node_line_1")
		os.makedirs(node_line_path)
		yield node_line_path


def check_best_result(best_df: pd.DataFrame):
	assert isinstance(best_df, pd.DataFrame)
	assert len(best_df) == len(previous_result)
	assert set(best_df.columns) == {
		"query",
		"retrieved_contents",
		"test_column",
		"prompts",
		"prompt_maker_bleu",
		"prompt_maker_rouge",
	}


def check_summary_df(node_line_dir):
	# check the files saved properly
	summary_path = os.path.join(node_line_dir, "prompt_maker", "summary.csv")
	assert os.path.exists(summary_path)
	summary_df = load_summary_file(summary_path)
	assert len(summary_df) == len(previous_result)
	assert set(summary_df.columns) == {
		"filename",
		"module_name",
		"module_params",
		"execution_time",
		"average_prompt_token",
		"prompt_maker_bleu",
		"prompt_maker_rouge",
		"is_best",
	}
	best_filename = summary_df[summary_df["is_best"]]["filename"].values[0]
	return best_filename


def test_run_prompt_maker_node(node_line_dir):
	generator_models["mock"] = MockLLM
	modules = [Fstring, Fstring]
	params = [
		{
			"prompt": "Tell me something about the question: {query} \n\n {retrieved_contents}"
		},
		{
			"prompt": "Question: {query} \n Something to read: {retrieved_contents} \n What's your answer?"
		},
	]
	strategies = {
		"metrics": metrics,
		"speed_threshold": 5,
		"token_threshold": 25,
		"tokenizer": "gpt-3.5-turbo",
		"strategy": "rank",
		"generator_modules": [
			{
				"module_type": "llama_index_llm",
				"llm": "mock",
				"model": ["gpt-3.5-turbo", "gpt-3.5-turbo-1106"],
			}
		],
	}
	best_result = run_prompt_maker_node(
		modules, params, previous_result, node_line_dir, strategies
	)
	check_best_result(best_result)
	best_filename = check_summary_df(node_line_dir)
	assert best_filename == "0.parquet"
	best_result_path = os.path.join(
		node_line_dir, "prompt_maker", f"best_{best_filename}"
	)
	assert os.path.exists(best_result_path)

	assert os.path.exists(os.path.join(node_line_dir, "prompt_maker", "0.parquet"))
	assert os.path.exists(os.path.join(node_line_dir, "prompt_maker", "1.parquet"))


async def acomplete_qa_creation(*args, **kwargs):
	return CompletionResponse(text="This is the test answer.")


@patch.object(OpenAI, "acomplete", acomplete_qa_creation)
def test_run_prompt_maker_node_default(node_line_dir):
	modules = [Fstring, Fstring]
	params = [
		{
			"prompt": "Tell me something about the question: {query} \n\n {retrieved_contents}"
		},
		{
			"prompt": "Question: {query} \n Something to read: {retrieved_contents} \n What's your answer?"
		},
	]
	strategies = {"metrics": metrics}
	best_result = run_prompt_maker_node(
		modules, params, previous_result, node_line_dir, strategies
	)
	check_best_result(best_result)
	best_filename = check_summary_df(node_line_dir)
	best_result_path = os.path.join(
		node_line_dir, "prompt_maker", f"best_{best_filename}"
	)
	assert os.path.exists(best_result_path)


def test_run_prompt_maker_one_module(node_line_dir):
	modules = [Fstring]
	params = [
		{
			"prompt": "Tell me something about the question: {query} \n\n {retrieved_contents}"
		}
	]
	strategies = {"metrics": metrics}
	best_result = run_prompt_maker_node(
		modules, params, previous_result, node_line_dir, strategies
	)
	assert set(best_result.columns) == {
		"query",
		"retrieved_contents",
		"test_column",
		"prompts",  # automatically skip evaluation
	}
	summary_filepath = os.path.join(node_line_dir, "prompt_maker", "summary.csv")
	assert os.path.exists(summary_filepath)
	summary_df = load_summary_file(summary_filepath)
	assert set(summary_df) == {
		"filename",
		"module_name",
		"module_params",
		"execution_time",
		"is_best",
		"average_prompt_token",
	}
	best_filepath = os.path.join(
		node_line_dir, "prompt_maker", f"best_{summary_df['filename'].values[0]}"
	)
	assert os.path.exists(best_filepath)


def test_summary_metric_name_at_threshold_cutoff():
	evaluation_results = [
		pd.DataFrame(
			{
				"test_column": ["test_value_1", "test_value_2", "test_value_3"],
				"bleu": [0.1, 0.2, 0.3],
				"rouge": [0.4, 0.5, 0.6],
			}
		)
	]
	summary_df = pd.DataFrame(
		{
			"filename": ["filename_1", "filename_2"],
			"execution_time": [1, 2],
		}
	)
	filenames = ["filename_2"]
	metric_names, _ = cast_metrics(metrics)
	evaluation_df = pd.DataFrame(
		{
			"filename": filenames,
			**{
				f"prompt_maker_{metric_name}": list(
					map(lambda x: x[metric_name].mean(), evaluation_results)
				)
				for metric_name in metric_names
			},
		}
	)
	summary_df = pd.merge(
		on="filename", left=summary_df, right=evaluation_df, how="left"
	)

	assert set(summary_df.columns) == {
		"filename",
		"execution_time",
		"prompt_maker_bleu",
		"prompt_maker_rouge",
	}
	assert np.isnan(summary_df["prompt_maker_bleu"].values[0])
	assert np.isnan(summary_df["prompt_maker_rouge"].values[0])
	assert summary_df["prompt_maker_bleu"].values[1] == pytest.approx(0.2)
	assert summary_df["prompt_maker_rouge"].values[1] == pytest.approx(0.5)



import os.path
import tempfile

from autorag.data.parse.table_hybrid_parse import table_hybrid_parse
from autorag.data.parse.table_hybrid_parse import save_page_by_table, get_page_from_path

from tests.autorag.data.parse.test_parse_base import (
	hybrid_data_list,
	hybrid_glob,
	check_parse_result,
	multiple_pdf_data_list,
	eng_text_glob,
	table_data_list,
	korean_table_glob,
)

table_hybrid_params = {
	"text_parse_module": "langchain_parse",
	"text_params": {"parse_method": "pdfminer"},
	"table_parse_module": "langchain_parse",
	"table_params": {"parse_method": "pdfplumber"},
}


def test_table_hybrid_parse():
	table_hybrid_parse_original = table_hybrid_parse.__wrapped__
	texts, path, pages = table_hybrid_parse_original(
		hybrid_data_list, **table_hybrid_params
	)
	check_parse_result(texts, path, pages, "hybrid", "hybrid")


def test_table_hybrid_parse_node():
	result_df = table_hybrid_parse(hybrid_glob, **table_hybrid_params)
	check_parse_result(
		result_df["texts"].tolist(),
		result_df["path"].tolist(),
		result_df["page"].tolist(),
		"hybrid",
		"hybrid",
	)


def test_table_hybrid_parse_only_text():
	table_hybrid_parse_original = table_hybrid_parse.__wrapped__
	texts, path, pages = table_hybrid_parse_original(
		multiple_pdf_data_list, **table_hybrid_params
	)
	check_parse_result(texts, path, pages, "hybrid_text", "hybrid")


def test_table_hybrid_parse_only_text_node():
	result_df = table_hybrid_parse(eng_text_glob, **table_hybrid_params)
	check_parse_result(
		result_df["texts"].tolist(),
		result_df["path"].tolist(),
		result_df["page"].tolist(),
		"hybrid_text",
		"hybrid",
	)


def test_table_hybrid_parse_only_table():
	table_hybrid_parse_original = table_hybrid_parse.__wrapped__
	texts, path, pages = table_hybrid_parse_original(
		table_data_list, **table_hybrid_params
	)
	check_parse_result(texts, path, pages, "hybrid_table", "hybrid")


def test_table_hybrid_parse_only_table_node():
	result_df = table_hybrid_parse(korean_table_glob, **table_hybrid_params)
	check_parse_result(
		result_df["texts"].tolist(),
		result_df["path"].tolist(),
		result_df["page"].tolist(),
		"hybrid_table",
		"hybrid",
	)


def test_save_page_by_table():
	with tempfile.TemporaryDirectory(ignore_cleanup_errors=True) as save_dir:
		text_dir = os.path.join(save_dir, "text")
		table_dir = os.path.join(save_dir, "table")
		os.makedirs(text_dir, exist_ok=True)
		os.makedirs(table_dir, exist_ok=True)

		path_map_dict = save_page_by_table(hybrid_data_list[0], text_dir, table_dir)
		text_page_dir = os.path.join(text_dir, "nfl_rulebook_both_page_1.pdf")
		table_page_dir = os.path.join(table_dir, "nfl_rulebook_both_page_2.pdf")
		assert os.path.exists(text_dir)
		assert os.path.exists(text_page_dir)
		assert os.path.exists(table_dir)
		assert os.path.exists(table_page_dir)
		assert path_map_dict == {
			text_page_dir: hybrid_data_list[0],
			table_page_dir: hybrid_data_list[0],
		}


def test_split_name_page():
	file_path = "havertz/jax/jeffrey_love_story_page_1.pdf"
	page_num = get_page_from_path(file_path)
	assert page_num == 1


def test_split_name_page_two_page():
	file_path = "havertz/jax/jeffrey_love_story_bali_page_page_2.pdf"
	page_num = get_page_from_path(file_path)
	assert page_num == 2



import math
from typing import Tuple, List

import pandas as pd

from autorag.evaluation import evaluate_retrieval
from autorag.schema.metricinput import MetricInput

retrieval_gt = [[[f"test{i}-{j}"] for i in range(2)] for j in range(4)]
queries_example = ["Query 1", "Query 2", "Query 3", "Query 4"]
generation_gt_example = [
	["Jazz", "Music"],
	["Havertz"],
	["William", "Gamst"],
	["Kia", "Tigers", "V12"],
]
contents = [
	["a", "b", "c", "d"],
	["better", "bone", "caleb", "done"],
	["ate", "better", "cortex", "dad"],
	["aim", "bond", "cane", "dance"],
]
scores = [
	[0.1, 0.2, 0.3, 0.4],
	[0.4, 0.3, 0.2, 0.1],
	[0.5, 0.4, 0.3, 0.2],
	[0.6, 0.5, 0.4, 0.3],
]
ids = [
	[retrieval_gt[0][0][0], retrieval_gt[0][1][0], "pred-0", "pred-1"],
	["pred-2", retrieval_gt[1][0][0], retrieval_gt[1][1][0], "pred-3"],
	[f"pred-{i}" for i in range(4, 8)],
	[retrieval_gt[3][0][0], "pred-8", "pred-9", "pred-10"],
]
metric_inputs = [MetricInput(retrieval_gt=ret_gt, queries=queries, generation_gt=gen_gt) for ret_gt, queries, gen_gt in
				 zip(retrieval_gt, queries_example, generation_gt_example)]


@evaluate_retrieval(
	metric_inputs=metric_inputs,
	metrics=["retrieval_recall", "retrieval_precision", "retrieval_f1"],
)
def pseudo_retrieval() -> Tuple[List[List[str]], List[List[str]], List[List[float]]]:
	return contents, ids, scores


@evaluate_retrieval(
	metric_inputs=metric_inputs,
	metrics=[{"metric_name": "retrieval_recall"}, {"metric_name": "retrieval_f1"}],
)
def pseudo_retrieval_dict_metric() -> (
	Tuple[List[List[str]], List[List[str]], List[List[float]]]
):
	return contents, ids, scores


def test_evaluate_retrieval():
	result_df = pseudo_retrieval()
	assert isinstance(result_df, pd.DataFrame)
	assert len(result_df) == 4
	assert len(result_df.columns) == 6
	assert list(result_df.columns) == [
		"retrieved_contents",
		"retrieved_ids",
		"retrieve_scores",
		"retrieval_recall",
		"retrieval_precision",
		"retrieval_f1",
	]
	recall = result_df["retrieval_recall"].tolist()
	recall_solution = [1, 1, 0, 0.5]
	for pred, sol in zip(recall, recall_solution):
		assert math.isclose(pred, sol, rel_tol=1e-4)

	precision = result_df["retrieval_precision"].tolist()
	precision_solution = [0.5, 0.5, 0, 0.25]
	for pred, sol in zip(precision, precision_solution):
		assert math.isclose(pred, sol, rel_tol=1e-4)

	f1 = result_df["retrieval_f1"].tolist()
	f1_solution = [1 / 1.5, 1 / 1.5, 0, 0.25 / 0.75]
	for pred, sol in zip(f1, f1_solution):
		assert math.isclose(pred, sol, rel_tol=1e-4)


def test_evaluate_retrieval_dict():
	result_df = pseudo_retrieval_dict_metric()
	assert isinstance(result_df, pd.DataFrame)
	assert len(result_df) == 4
	assert len(result_df.columns) == 5
	assert set(result_df.columns) == {
		"retrieved_contents",
		"retrieved_ids",
		"retrieve_scores",
		"retrieval_recall",
		"retrieval_f1",
	}

	recall = result_df["retrieval_recall"].tolist()
	recall_solution = [1, 1, 0, 0.5]
	for pred, sol in zip(recall, recall_solution):
		assert math.isclose(pred, sol, rel_tol=1e-4)



from autorag.nodes.retrieval import BM25, VectorDB
from autorag.nodes.retrieval.run import run_retrieval_node
from autorag.schema import Node
from autorag.schema.module import Module
from autorag.schema.node import extract_values_from_nodes, module_type_exists


def test_get_param_combinations():
	modules = [
		Module.from_dict(
			{
				"module_type": "bm25",
				"key2": ["value1", "value2"],
				"key3": "value3",
				"key4": ["value4", "value5"],
			}
		),
		Module.from_dict(
			{
				"module_type": "vectordb",
				"key5": ["value6", "value6", "value7"],
				"key6": "value8",
			}
		),
	]
	node = Node(
		node_type="retrieval",
		strategy={"strategy_key": "strategy_value"},
		node_params={"param1": "value1"},
		modules=modules,
	)

	modules, module_node_params = node.get_param_combinations()

	assert isinstance(module_node_params, list)
	assert isinstance(modules, list)
	assert len(modules) == 6
	assert modules == [BM25, BM25, BM25, BM25, VectorDB, VectorDB]
	bm25_solution = [
		{"param1": "value1", "key2": "value1", "key3": "value3", "key4": "value4"},
		{"param1": "value1", "key2": "value1", "key3": "value3", "key4": "value5"},
		{"param1": "value1", "key2": "value2", "key3": "value3", "key4": "value4"},
		{"param1": "value1", "key2": "value2", "key3": "value3", "key4": "value5"},
	]
	vectordb_solution = [
		{"param1": "value1", "key5": "value6", "key6": "value8"},
		{"param1": "value1", "key5": "value7", "key6": "value8"},
	]
	for module, module_params in zip(modules, module_node_params):
		if module == BM25:
			assert module_params in bm25_solution
			bm25_solution.pop(bm25_solution.index(module_params))
		elif module == VectorDB:
			assert module_params in vectordb_solution
			vectordb_solution.pop(vectordb_solution.index(module_params))
		else:
			raise ValueError(f"Module {module} is not supposed to be here.")
	assert len(bm25_solution) == 0
	assert len(vectordb_solution) == 0


def test_from_dict():
	node_dict = {
		"node_type": "retrieval",
		"strategy": {"strategy_key": "strategy_value"},
		"modules": [{"module_type": "bm25"}, {"module_type": "bm25", "key2": "value2"}],
		"extra_param": "extra_value",
	}

	node = Node.from_dict(node_dict)

	assert node.node_type == "retrieval"
	assert node.run_node == run_retrieval_node
	assert node.strategy == {"strategy_key": "strategy_value"}
	assert node.node_params == {"extra_param": "extra_value"}
	assert all(isinstance(mod, Module) for mod in node.modules)
	assert len(node.modules) == 2
	assert node.modules[0].module_param == {}
	assert node.modules[1].module_param == {"key2": "value2"}


def test_find_embedding_models():
	nodes = [
		Node.from_dict(
			{
				"node_type": "retrieval",
				"param1": "value1",
				"strategy": {
					"metrics": ["retrieval_f1", "retrieval_recall"],
				},
				"modules": [
					{"module_type": "bm25"},
					{
						"module_type": "bm25",
						"param2": "value2",
						"embedding_model": ["model1", "model2"],
					},
					{
						"module_type": "bm25",
						"param2": "value3",
						"param3": ["value4", "value5"],
						"embedding_model": ["model1", "model3"],
					},
				],
			}
		),
		Node.from_dict(
			{
				"node_type": "retrieval",
				"strategy": {
					"metrics": ["retrieval_f1"],
				},
				"modules": [
					{
						"module_type": "bm25",
						"param2": "value2",
						"embedding_model": ["model1", "model3", "model4"],
					}
				],
			}
		),
	]
	embedding_models = extract_values_from_nodes(nodes, "embedding_model")
	assert set(embedding_models) == {"model1", "model2", "model3", "model4"}


def test_find_llm_models():
	nodes = [
		Node.from_dict(
			{
				"node_type": "retrieval",
				"param1": "value1",
				"strategy": {
					"metrics": ["retrieval_f1", "retrieval_recall"],
				},
				"modules": [
					{"module_type": "bm25"},
					{
						"module_type": "bm25",
						"param2": "value2",
						"llm": "model1",
					},
					{
						"module_type": "bm25",
						"param2": "value3",
						"param3": ["value4", "value5"],
						"llm": ["model1", "model3"],
					},
				],
			}
		),
		Node.from_dict(
			{
				"node_type": "retrieval",
				"strategy": {
					"metrics": ["retrieval_f1"],
				},
				"modules": [
					{
						"module_type": "bm25",
						"param2": "value2",
						"llm": ["model1", "model2", "model4"],
					}
				],
			}
		),
	]
	llm_models = extract_values_from_nodes(nodes, "llm")
	assert set(llm_models) == {"model1", "model2", "model3", "model4"}
	assert module_type_exists(nodes, "bm25") is True
	assert module_type_exists(nodes, "bm26") is False



from itertools import chain
from typing import List, Tuple

import pandas as pd
import torch
from tqdm import tqdm
from transformers import T5Tokenizer, T5ForConditionalGeneration

from autorag.nodes.passagereranker.base import BasePassageReranker
from autorag.utils.util import (
	make_batch,
	sort_by_scores,
	flatten_apply,
	select_top_k,
	result_to_dataframe,
	pop_params,
)

prediction_tokens = {
	"castorini/monot5-base-msmarco": ["▁false", "▁true"],
	"castorini/monot5-base-msmarco-10k": ["▁false", "▁true"],
	"castorini/monot5-large-msmarco": ["▁false", "▁true"],
	"castorini/monot5-large-msmarco-10k": ["▁false", "▁true"],
	"castorini/monot5-base-med-msmarco": ["▁false", "▁true"],
	"castorini/monot5-3b-med-msmarco": ["▁false", "▁true"],
	"castorini/monot5-3b-msmarco-10k": ["▁false", "▁true"],
	"unicamp-dl/mt5-base-en-msmarco": ["▁no", "▁yes"],
	"unicamp-dl/ptt5-base-pt-msmarco-10k-v2": ["▁não", "▁sim"],
	"unicamp-dl/ptt5-base-pt-msmarco-100k-v2": ["▁não", "▁sim"],
	"unicamp-dl/ptt5-base-en-pt-msmarco-100k-v2": ["▁não", "▁sim"],
	"unicamp-dl/mt5-base-en-pt-msmarco-v2": ["▁no", "▁yes"],
	"unicamp-dl/mt5-base-mmarco-v2": ["▁no", "▁yes"],
	"unicamp-dl/mt5-base-en-pt-msmarco-v1": ["▁no", "▁yes"],
	"unicamp-dl/mt5-base-mmarco-v1": ["▁no", "▁yes"],
	"unicamp-dl/ptt5-base-pt-msmarco-10k-v1": ["▁não", "▁sim"],
	"unicamp-dl/ptt5-base-pt-msmarco-100k-v1": ["▁não", "▁sim"],
	"unicamp-dl/ptt5-base-en-pt-msmarco-10k-v1": ["▁não", "▁sim"],
	"unicamp-dl/mt5-3B-mmarco-en-pt": ["▁", "▁true"],
	"unicamp-dl/mt5-13b-mmarco-100k": ["▁", "▁true"],
}


class MonoT5(BasePassageReranker):
	def __init__(
		self,
		project_dir: str,
		model_name: str = "castorini/monot5-3b-msmarco-10k",
		*args,
		**kwargs,
	):
		"""
		Initialize the MonoT5 reranker.

		:param project_dir: The project directory
		:param model_name: The name of the MonoT5 model to use for reranking
			Note: default model name is 'castorini/monot5-3b-msmarco-10k'
				If there is a '/' in the model name parameter,
				when we create the file to store the results, the path will be twisted because of the '/'.
				Therefore, it will be received as '_' instead of '/'.
		:param kwargs: The extra arguments for the MonoT5 reranker
		"""
		super().__init__(project_dir)
		# replace '_' to '/'
		if "_" in model_name:
			model_name = model_name.replace("_", "/")
		# Load the tokenizer and model from the pre-trained MonoT5 model
		self.tokenizer = T5Tokenizer.from_pretrained(model_name)
		model_params = pop_params(T5ForConditionalGeneration.from_pretrained, kwargs)
		self.model = T5ForConditionalGeneration.from_pretrained(
			model_name, **model_params
		).eval()

		# Determine the device to run the model on (GPU if available, otherwise CPU)
		self.device = "cuda" if torch.cuda.is_available() else "cpu"
		self.model.to(self.device)

		token_false, token_true = prediction_tokens[model_name]
		self.token_false_id = self.tokenizer.convert_tokens_to_ids(token_false)
		self.token_true_id = self.tokenizer.convert_tokens_to_ids(token_true)

	@result_to_dataframe(["retrieved_contents", "retrieved_ids", "retrieve_scores"])
	def pure(self, previous_result: pd.DataFrame, *args, **kwargs):
		queries, contents, _, ids = self.cast_to_run(previous_result)
		top_k = kwargs.get("top_k", 3)
		batch = kwargs.get("batch", 64)
		return self._pure(queries, contents, ids, top_k, batch)

	def _pure(
		self,
		queries: List[str],
		contents_list: List[List[str]],
		ids_list: List[List[str]],
		top_k: int,
		batch: int = 64,
	) -> Tuple[List[List[str]], List[List[str]], List[List[float]]]:
		"""
		Rerank a list of contents based on their relevance to a query using MonoT5.

		:param queries: The list of queries to use for reranking
		:param contents_list: The list of lists of contents to rerank
		:param ids_list: The list of lists of ids retrieved from the initial ranking
		:param top_k: The number of passages to be retrieved

		:param batch: The number of queries to be processed in a batch
		:return: tuple of lists containing the reranked contents, ids, and scores
		"""
		# Retrieve the tokens used by the model to represent false and true predictions

		nested_list = [
			list(map(lambda x: [f"Query: {query} Document: {x}"], content_list))
			for query, content_list in zip(queries, contents_list)
		]

		rerank_scores = flatten_apply(
			monot5_run_model,
			nested_list,
			model=self.model,
			batch_size=batch,
			tokenizer=self.tokenizer,
			device=self.device,
			token_false_id=self.token_false_id,
			token_true_id=self.token_true_id,
		)

		df = pd.DataFrame(
			{
				"contents": contents_list,
				"ids": ids_list,
				"scores": rerank_scores,
			}
		)
		df[["contents", "ids", "scores"]] = df.apply(
			sort_by_scores, axis=1, result_type="expand"
		)
		results = select_top_k(df, ["contents", "ids", "scores"], top_k)

		return (
			results["contents"].tolist(),
			results["ids"].tolist(),
			results["scores"].tolist(),
		)


def monot5_run_model(
	input_texts,
	model,
	batch_size: int,
	tokenizer,
	device,
	token_false_id,
	token_true_id,
):
	batch_input_texts = make_batch(input_texts, batch_size)
	results = []
	for batch_texts in tqdm(batch_input_texts):
		flattened_batch_texts = list(chain.from_iterable(batch_texts))
		input_encodings = tokenizer(
			flattened_batch_texts,
			padding=True,
			truncation=True,
			max_length=512,
			return_tensors="pt",
		).to(device)
		with torch.no_grad():
			outputs = model.generate(
				input_ids=input_encodings["input_ids"],
				attention_mask=input_encodings["attention_mask"],
				output_scores=True,
				return_dict_in_generate=True,
			)

		# Extract logits for the 'false' and 'true' tokens from the model's output
		logits = outputs.scores[-1][:, [token_false_id, token_true_id]]
		# Calculate the softmax probability of the 'true' token
		probs = torch.nn.functional.softmax(logits, dim=-1)[:, 1]
		results.extend(probs.tolist())
	return results



import os
import pathlib
import re
import tempfile
from typing import Any
from unittest.mock import patch

import chromadb
import pandas as pd
import pytest
from llama_index.core.base.llms.types import CompletionResponse
from llama_index.embeddings.openai import OpenAIEmbedding

from autorag.data.legacy.qacreation import (
	make_single_content_qa,
	generate_qa_llama_index,
	make_qa_with_existing_qa,
	generate_answers,
)
from autorag.utils import validate_qa_dataset
from tests.mock import MockLLM, mock_get_text_embedding_batch

root_dir = pathlib.PurePath(
	os.path.dirname(os.path.realpath(__file__))
).parent.parent.parent.parent
resource_dir = os.path.join(root_dir, "resources")


@pytest.fixture
def qa_parquet_filepath():
	with tempfile.NamedTemporaryFile(suffix=".parquet", delete=False) as f:
		yield f.name
		f.close()
		os.unlink(f.name)


@pytest.fixture
def chroma_persistent_client():
	with tempfile.TemporaryDirectory(ignore_cleanup_errors=True) as temp_dir:
		client = chromadb.PersistentClient(temp_dir)
		yield client


async def acomplete_qa_creation(self, messages, **kwargs: Any):
	pattern = r"Output with (\d+) QnAs:"
	matches = re.findall(pattern, messages)
	num_questions = int(matches[-1])
	return CompletionResponse(
		text="[Q]: Is this the test question?\n[A]: Yes, this is the test answer."
		* num_questions
	)


@patch.object(
	MockLLM,
	"acomplete",
	acomplete_qa_creation,
)
def test_single_content_qa(qa_parquet_filepath):
	corpus_df = pd.read_parquet(
		os.path.join(resource_dir, "corpus_data_sample.parquet")
	)
	qa_df = make_single_content_qa(
		corpus_df,
		content_size=3,
		qa_creation_func=generate_qa_llama_index,
		output_filepath=qa_parquet_filepath,
		llm=MockLLM(),
		question_num_per_content=2,
		upsert=True,
	)
	validate_qa_dataset(qa_df)
	assert len(qa_df) == qa_df["qid"].nunique()
	assert len(qa_df) == 6
	assert qa_df["retrieval_gt"].tolist()[0] == qa_df["retrieval_gt"].tolist()[1]

	assert all([len(x) == 1 and len(x[0]) == 1 for x in qa_df["retrieval_gt"].tolist()])
	assert all([len(x) == 1 for x in qa_df["generation_gt"].tolist()])


@patch.object(
	MockLLM,
	"acomplete",
	acomplete_qa_creation,
)
def test_single_content_qa_long_cache_batch(qa_parquet_filepath):
	corpus_df = pd.read_parquet(
		os.path.join(resource_dir, "corpus_data_sample.parquet")
	)
	qa_df = make_single_content_qa(
		corpus_df,
		content_size=30,
		qa_creation_func=generate_qa_llama_index,
		output_filepath=qa_parquet_filepath,
		llm=MockLLM(),
		question_num_per_content=2,
		upsert=True,
		cache_batch=2,
	)
	validate_qa_dataset(qa_df)
	assert len(qa_df) == 60
	assert qa_df["retrieval_gt"].tolist()[0] == qa_df["retrieval_gt"].tolist()[1]

	assert all([len(x) == 1 and len(x[0]) == 1 for x in qa_df["retrieval_gt"].tolist()])
	assert all([len(x) == 1 for x in qa_df["generation_gt"].tolist()])


@patch.object(
	OpenAIEmbedding,
	"get_text_embedding_batch",
	mock_get_text_embedding_batch,
)
def test_make_qa_with_existing_qa_without_gen_gt(qa_parquet_filepath):
	corpus_df = pd.read_parquet(
		os.path.join(resource_dir, "corpus_data_sample.parquet"), engine="pyarrow"
	)
	query_df = pd.read_parquet(
		os.path.join(resource_dir, "qa_data_sample.parquet"), engine="pyarrow"
	)
	qa_df = make_qa_with_existing_qa(
		corpus_df,
		query_df,
		content_size=5,
		answer_creation_func=generate_answers,
		output_filepath=qa_parquet_filepath,
		llm=MockLLM(),
		upsert=True,
	)
	validate_qa_dataset(qa_df)
	assert len(qa_df) == 5
	assert all(
		len(elem) == 3 for elem in qa_df["retrieval_gt"].apply(lambda x: x[0]).tolist()
	)
	assert all((elem in query_df["query"].tolist()) for elem in qa_df["query"].tolist())


@patch.object(
	OpenAIEmbedding,
	"get_text_embedding_batch",
	mock_get_text_embedding_batch,
)
def test_make_qa_with_existing_qa_with_gen_gt(qa_parquet_filepath):
	corpus_df = pd.read_parquet(
		os.path.join(resource_dir, "corpus_data_sample.parquet"), engine="pyarrow"
	)
	query_df = pd.read_parquet(
		os.path.join(resource_dir, "qa_data_sample.parquet"), engine="pyarrow"
	)
	qa_df = make_qa_with_existing_qa(
		corpus_df,
		query_df,
		content_size=5,
		exist_gen_gt=True,
		output_filepath=qa_parquet_filepath,
		upsert=True,
	)
	validate_qa_dataset(qa_df)
	assert len(qa_df) == 5
	assert all(
		len(elem) == 3 for elem in qa_df["retrieval_gt"].apply(lambda x: x[0]).tolist()
	)
	assert all((elem in query_df["query"].tolist()) for elem in qa_df["query"].tolist())
	assert all(
		(elem in query_df["generation_gt"].tolist())
		for elem in qa_df["generation_gt"].tolist()
	)


@patch.object(
	OpenAIEmbedding,
	"get_text_embedding_batch",
	mock_get_text_embedding_batch,
)
def test_make_qa_with_existing_qa_persistent_client_without_gen_gt(
	chroma_persistent_client, qa_parquet_filepath
):
	corpus_df = pd.read_parquet(
		os.path.join(resource_dir, "corpus_data_sample.parquet"), engine="pyarrow"
	)
	query_df = pd.read_parquet(
		os.path.join(resource_dir, "qa_data_sample.parquet"), engine="pyarrow"
	)
	collection = chroma_persistent_client.get_or_create_collection("auto-rag")
	qa_df = make_qa_with_existing_qa(
		corpus_df,
		query_df,
		content_size=5,
		answer_creation_func=generate_answers,
		output_filepath=qa_parquet_filepath,
		llm=MockLLM(),
		upsert=True,
		collection=collection,
		embedding_model="openai_embed_3_small",
	)
	validate_qa_dataset(qa_df)
	assert len(qa_df) == 5
	assert all(
		len(elem) == 3 for elem in qa_df["retrieval_gt"].apply(lambda x: x[0]).tolist()
	)
	assert all((elem in query_df["query"].tolist()) for elem in qa_df["query"].tolist())


@patch.object(
	OpenAIEmbedding,
	"get_text_embedding_batch",
	mock_get_text_embedding_batch,
)
def test_make_qa_with_existing_qa_persistent_client_with_gen_gt(
	chroma_persistent_client, qa_parquet_filepath
):
	corpus_df = pd.read_parquet(
		os.path.join(resource_dir, "corpus_data_sample.parquet"), engine="pyarrow"
	)
	query_df = pd.read_parquet(
		os.path.join(resource_dir, "qa_data_sample.parquet"), engine="pyarrow"
	)
	collection = chroma_persistent_client.get_or_create_collection("auto-rag")
	qa_df = make_qa_with_existing_qa(
		corpus_df,
		query_df,
		content_size=5,
		exist_gen_gt=True,
		output_filepath=qa_parquet_filepath,
		upsert=True,
		collection=collection,
		embedding_model="openai_embed_3_small",
	)
	validate_qa_dataset(qa_df)
	assert len(qa_df) == 5
	assert all(
		len(elem) == 3 for elem in qa_df["retrieval_gt"].apply(lambda x: x[0]).tolist()
	)
	assert all((elem in query_df["query"].tolist()) for elem in qa_df["query"].tolist())
	assert all(
		(elem in query_df["generation_gt"].tolist())
		for elem in qa_df["generation_gt"].tolist()
	)



from unittest.mock import patch

import pandas as pd
import pytest
from llama_index.llms.openai import OpenAI

from autorag import generator_models
from autorag.nodes.passagecompressor import TreeSummarize
from tests.autorag.nodes.passagecompressor.test_base_passage_compressor import (
	queries,
	retrieved_contents,
	df,
	check_result,
)
from tests.mock import MockLLM


@pytest.fixture
def tree_summarize_instance():
	return TreeSummarize("project_dir", llm="mock")


@pytest.fixture
def tree_summarize_instance_chat():
	return TreeSummarize("project_dir", llm="openai", model="gpt-4o-mini")


async def mock_openai_apredict(self, prompt, *args, **kwargs):
	return "What is the capital of France?"


def test_tree_summarize_default(tree_summarize_instance):
	result = tree_summarize_instance._pure(queries, retrieved_contents)
	check_result(result)


@patch.object(
	OpenAI,
	"apredict",
	mock_openai_apredict,
)
def test_tree_summarize_chat(tree_summarize_instance_chat):
	result = tree_summarize_instance_chat._pure(queries, retrieved_contents)
	check_result(result)


def test_tree_summarize_custom_prompt(tree_summarize_instance):
	prompt = "This is a custom prompt. {context_str} {query_str}"
	result = tree_summarize_instance._pure(
		queries,
		retrieved_contents,
		prompt=prompt,
	)
	check_result(result)
	assert bool(result[0]) is True


@patch.object(
	OpenAI,
	"apredict",
	mock_openai_apredict,
)
def test_tree_summarize_custom_prompt_chat(tree_summarize_instance_chat):
	prompt = "Query: {query_str} Passages: {context_str}. Repeat the query."
	result = tree_summarize_instance_chat._pure(
		queries, retrieved_contents, chat_prompt=prompt
	)
	check_result(result)
	assert "What is the capital of France?" in result[0]


def test_tree_summarize_node():
	generator_models["mock"] = MockLLM
	result = TreeSummarize.run_evaluator(
		"project_dir",
		df,
		llm="mock",
		prompt="This is a custom prompt. {context_str} {query_str}",
		max_tokens=64,
	)
	assert isinstance(result, pd.DataFrame)
	contents = result["retrieved_contents"].tolist()
	assert isinstance(contents, list)
	assert len(contents) == len(queries)
	assert isinstance(contents[0], list)
	assert len(contents[0]) == 1
	assert isinstance(contents[0][0], str)
	assert bool(contents[0][0]) is True



from typing import List, Tuple

from llama_parse import LlamaParse

from autorag.data.parse.base import parser_node
from autorag.utils.util import process_batch, get_event_loop


@parser_node
def llama_parse(
	data_path_list: List[str], batch: int = 8, **kwargs
) -> Tuple[List[str], List[str], List[int]]:
	"""
	Parse documents to use llama_parse.
	LLAMA_CLOUD_API_KEY environment variable should be set.
	You can get the key from https://cloud.llamaindex.ai/api-key

	:param data_path_list: The list of data paths to parse.
	:param batch: The batch size for parse documents. Default is 8.
	:param kwargs: The extra parameters for creating the llama_parse instance.
	:return: tuple of lists containing the parsed texts, path and pages.
	"""
	parse_instance = LlamaParse(**kwargs)

	tasks = [
		llama_parse_pure(data_path, parse_instance) for data_path in data_path_list
	]
	loop = get_event_loop()
	results = loop.run_until_complete(process_batch(tasks, batch))

	del parse_instance

	texts, path = zip(*results)
	pages = [-1] * len(texts)

	return list(texts), list(path), pages


async def llama_parse_pure(data_path: str, parse_instance) -> Tuple[str, str]:
	documents = await parse_instance.aload_data(data_path)

	text = documents[0].text

	return text, data_path
